//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/InterferenceItems.cpp
//! @brief     Implements InterferenceItems's classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/InterferenceItems.h"
#include "Base/Const/Units.h"
#include "GUI/Model/Sample/FTDecayFunctionItems.h"
#include "GUI/Model/Sample/FTDistributionItems.h"
#include "GUI/Model/Sample/Lattice2DItems.h"
#include "GUI/Model/Types/UIntDescriptor.h"
#include "Sample/Aggregate/Interferences.h"

// TODO (when back compatibility will be broken  again)
// Make Interference1DLatticeItem::P_DECAY_FUNCTION and
// Interference2DLatticeItem::P_DECAY_FUNCTION rely on same constant

InterferenceItem::~InterferenceItem() = default;

DoubleDescriptor InterferenceItem::positionVariance() const
{
    DoubleDescriptor d(getItem(P_POSITION_VARIANCE), Unit::nanometerPower2);
    d.tooltip = "Variance of the position in each dimension";
    return d;
}

InterferenceItem::InterferenceItem(const QString& modelType) : SessionItem(modelType)
{
    addProperty(P_POSITION_VARIANCE, 0.0)
        ->setToolTip("Variance of the position in each dimension (nm^2)");
}

// --------------------------------------------------------------------------------------------- //

Interference1DLatticeItem::Interference1DLatticeItem() : InterferenceItem(M_TYPE)
{
    addProperty(P_LENGTH, 20.0 * Units::nm)->setToolTip("Lattice length in nanometers");
    addProperty(P_ROTATION_ANGLE, 0.0)
        ->setToolTip("Rotation of lattice with respect to x-axis of reference \n"
                     "frame (beam direction) in degrees ");

    GroupInfo info;
    info.add<FTDecayFunction1DCauchyItem>();
    info.add<FTDecayFunction1DGaussItem>();
    info.add<FTDecayFunction1DTriangleItem>();
    info.add<FTDecayFunction1DVoigtItem>();
    info.setDefaultType<FTDecayFunction1DCauchyItem>();
    addGroupProperty(P_DECAY_FUNCTION, info)
        ->setToolTip("One-dimensional decay function (finite size effects)");
}

std::unique_ptr<IInterference> Interference1DLatticeItem::createInterference() const
{
    auto result = std::make_unique<Interference1DLattice>(
        getItemValue(P_LENGTH).toDouble(),
        Units::deg2rad(getItemValue(P_ROTATION_ANGLE).toDouble()));
    auto* pdfItem = dynamic_cast<FTDecayFunction1DItem*>(
        getGroupItem(Interference1DLatticeItem::P_DECAY_FUNCTION));
    result->setDecayFunction(*pdfItem->createFTDecayFunction());
    result->setPositionVariance(positionVariance());
    return std::unique_ptr<IInterference>(result.release());
}

DoubleDescriptor Interference1DLatticeItem::length() const
{
    DoubleDescriptor d(getItem(P_LENGTH), Unit::nanometer);
    d.tooltip = "Lattice length";
    return d;
}

DoubleDescriptor Interference1DLatticeItem::rotationAngle() const
{
    DoubleDescriptor d(getItem(P_ROTATION_ANGLE), Unit::degree);
    d.tooltip = "Rotation of lattice with respect to x-axis of reference frame (beam direction)";
    return d;
}

SelectionDescriptor<FTDecayFunction1DItem*> Interference1DLatticeItem::decayFunction() const
{
    return SelectionDescriptor<FTDecayFunction1DItem*>(item<GroupItem>(P_DECAY_FUNCTION));
}

// --------------------------------------------------------------------------------------------- //

SelectionDescriptor<Lattice2DItem*> Interference2DAbstractLatticeItem::latticeType() const
{
    return SelectionDescriptor<Lattice2DItem*>(item<GroupItem>(P_LATTICE_TYPE));
}

bool Interference2DAbstractLatticeItem::xiIntegration() const
{
    return getItemValue(P_XI_INTEGRATION).toBool();
}

void Interference2DAbstractLatticeItem::setXiIntegration(bool xi_integration)
{
    setItemValue(P_XI_INTEGRATION, xi_integration);
}

Interference2DAbstractLatticeItem::Interference2DAbstractLatticeItem(const QString& modelType,
                                                                     bool xi_integration)
    : InterferenceItem(modelType)
{
    GroupInfo info;
    info.add<BasicLattice2DItem>();
    info.add<SquareLattice2DItem>();
    info.add<HexagonalLattice2DItem>();
    info.setDefaultType<HexagonalLattice2DItem>();
    addGroupProperty(P_LATTICE_TYPE, info)->setToolTip("Type of lattice");

    addProperty(P_XI_INTEGRATION, xi_integration)
        ->setToolTip("Enables/disables averaging over the lattice rotation angle.");

    mapper()->setOnPropertyChange([this](const QString& name) {
        if (name == P_XI_INTEGRATION && isTag(P_LATTICE_TYPE))
            update_rotation_availability();
    });

    mapper()->setOnChildPropertyChange([this](SessionItem* item, const QString&) {
        if (item->hasModelType<GroupItem>() && item->displayName() == P_LATTICE_TYPE)
            update_rotation_availability();
    });
}

//! Sets rotation property of the lattice enabled/disabled depending on integration flag.

void Interference2DAbstractLatticeItem::update_rotation_availability()
{
    auto* p_lattice_item = latticeType().currentItem();
    if (p_lattice_item) {
        auto* angle_item = p_lattice_item->latticeRotationAngleItem();
        angle_item->setEnabled(!getItemValue(P_XI_INTEGRATION).toBool());
    }
}

// --------------------------------------------------------------------------------------------- //

Interference2DLatticeItem::Interference2DLatticeItem()
    : Interference2DAbstractLatticeItem(M_TYPE, false)
{
    GroupInfo info;
    info.add<FTDecayFunction2DCauchyItem>();
    info.add<FTDecayFunction2DGaussItem>();
    info.add<FTDecayFunction2DVoigtItem>();
    info.setDefaultType<FTDecayFunction2DCauchyItem>();
    addGroupProperty(P_DECAY_FUNCTION, info)
        ->setToolTip("Two-dimensional decay function (finite size effects)");
}

std::unique_ptr<IInterference> Interference2DLatticeItem::createInterference() const
{
    Lattice2DItem* latticeItem = latticeType().currentItem();
    std::unique_ptr<Interference2DLattice> result(
        new Interference2DLattice(*latticeItem->createLattice()));

    auto& pdfItem = groupItem<FTDecayFunction2DItem>(P_DECAY_FUNCTION);
    result->setDecayFunction(*pdfItem.createFTDecayFunction());
    result->setIntegrationOverXi(xiIntegration());
    result->setPositionVariance(positionVariance());

    return std::unique_ptr<IInterference>(result.release());
}

SelectionDescriptor<FTDecayFunction2DItem*> Interference2DLatticeItem::decayFunction() const
{
    return SelectionDescriptor<FTDecayFunction2DItem*>(item<GroupItem>(P_DECAY_FUNCTION));
}

// --------------------------------------------------------------------------------------------- //

Interference2DParaCrystalItem::Interference2DParaCrystalItem()
    : Interference2DAbstractLatticeItem(M_TYPE, true)
{
    latticeType().currentItem()->latticeRotationAngleItem()->setEnabled(false);

    addProperty(P_DAMPING_LENGTH, 0.0)
        ->setToolTip("The damping (coherence) length of the paracrystal in nanometers");

    addProperty(P_DOMAIN_SIZE1, 20.0 * Units::micrometer)
        ->setToolTip("Size of the coherent domain along the first basis vector in nanometers");
    addProperty(P_DOMAIN_SIZE2, 20.0 * Units::micrometer)
        ->setToolTip("Size of the coherent domain along the second basis vector in nanometers");

    GroupInfo info;
    info.add<FTDistribution2DCauchyItem>();
    info.add<FTDistribution2DGaussItem>();
    info.add<FTDistribution2DGateItem>();
    info.add<FTDistribution2DConeItem>();
    info.add<FTDistribution2DVoigtItem>();
    info.setDefaultType<FTDistribution2DCauchyItem>();

    addGroupProperty(P_PDF1, info)
        ->setToolTip("Probability distribution in first lattice direction");
    addGroupProperty(P_PDF2, info)
        ->setToolTip("Probability distribution in second lattice direction");

    mapper()->setOnChildPropertyChange([this](SessionItem* item, const QString& property) {
        if (item->hasModelType<GroupItem>() && GroupItem::isItemsTagName(property) && isTag(P_PDF1)
            && isTag(P_PDF2))
            update_distribution_displaynames();
    });

    update_distribution_displaynames();
}

std::unique_ptr<IInterference> Interference2DParaCrystalItem::createInterference() const
{
    Lattice2DItem* latticeItem = latticeType().currentItem();

    std::unique_ptr<Interference2DParaCrystal> result(
        new Interference2DParaCrystal(*latticeItem->createLattice(), 0, 0, 0));

    result->setDampingLength(getItemValue(P_DAMPING_LENGTH).toDouble());
    result->setDomainSizes(getItemValue(P_DOMAIN_SIZE1).toDouble(),
                           getItemValue(P_DOMAIN_SIZE2).toDouble());
    result->setIntegrationOverXi(xiIntegration());

    auto& pdf1Item = groupItem<FTDistribution2DItem>(Interference2DParaCrystalItem::P_PDF1);
    auto& pdf2Item = groupItem<FTDistribution2DItem>(Interference2DParaCrystalItem::P_PDF2);
    result->setProbabilityDistributions(*pdf1Item.createFTDistribution(),
                                        *pdf2Item.createFTDistribution());

    result->setPositionVariance(positionVariance());
    return std::unique_ptr<IInterference>(result.release());
}

DoubleDescriptor Interference2DParaCrystalItem::dampingLength() const
{
    DoubleDescriptor d(getItem(P_DAMPING_LENGTH), Unit::nanometer);
    d.tooltip = "The damping (coherence) length of the paracrystal";
    return d;
}

void Interference2DParaCrystalItem::setDampingLength(const double damping_length)
{
    setItemValue(P_DAMPING_LENGTH, damping_length);
}

DoubleDescriptor Interference2DParaCrystalItem::domainSize1() const
{
    DoubleDescriptor d(getItem(P_DOMAIN_SIZE1), Unit::nanometer);
    d.tooltip = "Size of the coherent domain along the first basis vector";
    return d;
}

void Interference2DParaCrystalItem::setDomainSize1(const double domain_size1)
{
    setItemValue(P_DOMAIN_SIZE1, domain_size1);
}

DoubleDescriptor Interference2DParaCrystalItem::domainSize2() const
{
    DoubleDescriptor d(getItem(P_DOMAIN_SIZE2), Unit::nanometer);
    d.tooltip = "Size of the coherent domain along the second basis vector";
    return d;
}

void Interference2DParaCrystalItem::setDomainSize2(const double domain_size2)
{
    setItemValue(P_DOMAIN_SIZE2, domain_size2);
}

SelectionDescriptor<FTDistribution2DItem*>
Interference2DParaCrystalItem::probabilityDistribution1() const
{
    return SelectionDescriptor<FTDistribution2DItem*>(item<GroupItem>(P_PDF1));
}

SelectionDescriptor<FTDistribution2DItem*>
Interference2DParaCrystalItem::probabilityDistribution2() const
{
    return SelectionDescriptor<FTDistribution2DItem*>(item<GroupItem>(P_PDF2));
}

void Interference2DParaCrystalItem::update_distribution_displaynames()
{
    auto* group1 = dynamic_cast<GroupItem*>(getItem(P_PDF1));
    auto* group2 = dynamic_cast<GroupItem*>(getItem(P_PDF2));

    if (!group1 || !group2)
        return;

    SessionItem* pdf1 = group1->currentItem();
    SessionItem* pdf2 = group2->currentItem();

    if (pdf1 && pdf2) {
        if (pdf1->modelType() == pdf2->modelType()) {
            pdf1->setDisplayName(pdf1->modelType() + QString::number(0));
            pdf2->setDisplayName(pdf2->modelType() + QString::number(1));
        } else {
            pdf1->setDisplayName(pdf1->modelType());
            pdf2->setDisplayName(pdf2->modelType());
        }
    }
}

// --------------------------------------------------------------------------------------------- //

InterferenceFinite2DLatticeItem::InterferenceFinite2DLatticeItem()
    : Interference2DAbstractLatticeItem(M_TYPE, false)
{
    addProperty(P_DOMAIN_SIZE_1, 100u)->setToolTip("Domain size 1 in number of unit cells");
    addProperty(P_DOMAIN_SIZE_2, 100u)->setToolTip("Domain size 2 in number of unit cells");
}

std::unique_ptr<IInterference> InterferenceFinite2DLatticeItem::createInterference() const
{
    Lattice2DItem* latticeItem = latticeType().currentItem();
    auto size_1 = getItemValue(P_DOMAIN_SIZE_1).toUInt();
    auto size_2 = getItemValue(P_DOMAIN_SIZE_2).toUInt();
    std::unique_ptr<InterferenceFinite2DLattice> result(
        new InterferenceFinite2DLattice(*latticeItem->createLattice(), size_1, size_2));

    result->setIntegrationOverXi(xiIntegration());
    result->setPositionVariance(positionVariance());

    return std::unique_ptr<IInterference>(result.release());
}

UIntDescriptor InterferenceFinite2DLatticeItem::domainSize1()
{
    return UIntDescriptor(getItem(P_DOMAIN_SIZE_1), Unit::unitless);
}

void InterferenceFinite2DLatticeItem::setDomainSize1(const unsigned int domain_size1)
{
    setItemValue(P_DOMAIN_SIZE_1, domain_size1);
}

UIntDescriptor InterferenceFinite2DLatticeItem::domainSize2()
{
    return UIntDescriptor(getItem(P_DOMAIN_SIZE_2), Unit::unitless);
}

void InterferenceFinite2DLatticeItem::setDomainSize2(const unsigned int domain_size2)
{
    setItemValue(P_DOMAIN_SIZE_2, domain_size2);
}

// --------------------------------------------------------------------------------------------- //

InterferenceHardDiskItem::InterferenceHardDiskItem() : InterferenceItem(M_TYPE)
{
    addProperty(P_RADIUS, 5.0 * Units::nm)->setToolTip("Hard disk radius in nanometers");
    addProperty(P_DENSITY, 0.002)->setToolTip("Particle density in particles per square nanometer");
}

std::unique_ptr<IInterference> InterferenceHardDiskItem::createInterference() const
{
    auto result = std::make_unique<InterferenceHardDisk>(getItemValue(P_RADIUS).toDouble(),
                                                         getItemValue(P_DENSITY).toDouble());
    result->setPositionVariance(positionVariance());
    return std::unique_ptr<IInterference>(result.release());
}

DoubleDescriptor InterferenceHardDiskItem::density() const
{
    DoubleDescriptor d(getItem(P_DENSITY), Unit::nanometerPowerMinus2);
    d.tooltip = "Particle density in particles per area";
    return d;
}

DoubleDescriptor InterferenceHardDiskItem::radius() const
{
    DoubleDescriptor d(getItem(P_RADIUS), Unit::nanometer);
    d.tooltip = "Hard disk radius";
    return d;
}

// --------------------------------------------------------------------------------------------- //

InterferenceRadialParaCrystalItem::InterferenceRadialParaCrystalItem() : InterferenceItem(M_TYPE)
{
    addProperty(P_PEAK_DISTANCE, 20.0 * Units::nm)
        ->setToolTip("Average distance to the next neighbor in nanometers");
    addProperty(P_DAMPING_LENGTH, 1000.0 * Units::nm)
        ->setToolTip("The damping (coherence) length of the paracrystal "
                     "in nanometers");
    addProperty(P_DOMAIN_SIZE, 0.0)
        ->setToolTip("Size of coherence domain along the lattice main axis "
                     "in nanometers");
    addProperty(P_KAPPA, 0.0)
        ->setToolTip("Size spacing coupling parameter of the Size Spacing "
                     "Correlation Approximation");

    GroupInfo info;
    info.add<FTDistribution1DCauchyItem>();
    info.add<FTDistribution1DGaussItem>();
    info.add<FTDistribution1DGateItem>();
    info.add<FTDistribution1DTriangleItem>();
    info.add<FTDistribution1DCosineItem>();
    info.add<FTDistribution1DVoigtItem>();
    info.setDefaultType<FTDistribution1DCauchyItem>();
    addGroupProperty(P_PDF, info)->setToolTip("One-dimensional probability distribution");
}

std::unique_ptr<IInterference> InterferenceRadialParaCrystalItem::createInterference() const
{
    auto result = std::make_unique<InterferenceRadialParaCrystal>(
        getItemValue(P_PEAK_DISTANCE).toDouble(), getItemValue(P_DAMPING_LENGTH).toDouble());
    result->setDomainSize(getItemValue(P_DOMAIN_SIZE).toDouble());
    result->setKappa(getItemValue(P_KAPPA).toDouble());

    auto& pdfItem = groupItem<FTDistribution1DItem>(P_PDF);
    result->setProbabilityDistribution(*pdfItem.createFTDistribution());
    result->setPositionVariance(positionVariance());
    return std::unique_ptr<IInterference>(result.release());
}

DoubleDescriptor InterferenceRadialParaCrystalItem::peakDistance() const
{
    DoubleDescriptor d(getItem(P_PEAK_DISTANCE), Unit::nanometer);
    d.tooltip = "Average distance to the next neighbor";
    return d;
}

DoubleDescriptor InterferenceRadialParaCrystalItem::dampingLength() const
{
    DoubleDescriptor d(getItem(P_DAMPING_LENGTH), Unit::nanometer);
    d.tooltip = "The damping (coherence) length of the paracrystal";
    return d;
}

DoubleDescriptor InterferenceRadialParaCrystalItem::domainSize() const
{
    DoubleDescriptor d(getItem(P_DOMAIN_SIZE), Unit::nanometer);
    d.tooltip = "Size of coherence domain along the lattice main axis";
    return d;
}

DoubleDescriptor InterferenceRadialParaCrystalItem::kappa() const
{
    return DoubleDescriptor(getItem(P_KAPPA), Unit::unitless);
}

SelectionDescriptor<FTDistribution1DItem*>
InterferenceRadialParaCrystalItem::probabilityDistribution() const
{
    return SelectionDescriptor<FTDistribution1DItem*>(item<GroupItem>(P_PDF));
}
