//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/Lattice2DItems.cpp
//! @brief     Implements classes Lattice2DItems
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/Lattice2DItems.h"
#include "Base/Const/Units.h"
#include "GUI/Model/Types/DoubleDescriptor.h"
#include "Sample/Lattice/Lattice2D.h"

namespace {

const QString axis_rotation_tooltip =
    "Rotation of lattice with respect to x-axis of reference frame \n"
    "(beam direction) in degrees";
}

Lattice2DItem::Lattice2DItem(const QString& modelType) : SessionItem(modelType)
{
    addProperty(P_LATTICE_ROTATION_ANGLE, 0.0)->setToolTip(axis_rotation_tooltip);
}

double Lattice2DItem::unitCellArea() const
{
    return createLattice()->unitCellArea();
}

SessionItem* Lattice2DItem::latticeRotationAngleItem() const
{
    return getItem(P_LATTICE_ROTATION_ANGLE);
}

DoubleDescriptor Lattice2DItem::latticeRotationAngle() const
{
    DoubleDescriptor d(getItem(P_LATTICE_ROTATION_ANGLE), Unit::degree);
    d.tooltip = "Rotation of lattice with respect to x-axis of reference frame (beam direction)";
    return d;
}

void Lattice2DItem::setLatticeRotationAngle(const double angle)
{
    setItemValue(P_LATTICE_ROTATION_ANGLE, angle);
}

// --------------------------------------------------------------------------------------------- //

BasicLattice2DItem::BasicLattice2DItem() : Lattice2DItem(M_TYPE)
{
    setToolTip("Two dimensional lattice");
    addProperty(P_LATTICE_LENGTH1, 20.0)
        ->setToolTip("Length of first lattice vector in nanometers");
    addProperty(P_LATTICE_LENGTH2, 20.0)
        ->setToolTip("Length of second lattice vector in nanometers");
    addProperty(P_LATTICE_ANGLE, 90.0)->setToolTip("Angle between lattice vectors in degrees");
}

std::unique_ptr<Lattice2D> BasicLattice2DItem::createLattice() const
{
    return std::make_unique<BasicLattice2D>(
        getItemValue(P_LATTICE_LENGTH1).toDouble(), getItemValue(P_LATTICE_LENGTH2).toDouble(),
        Units::deg2rad(getItemValue(P_LATTICE_ANGLE).toDouble()),
        Units::deg2rad(latticeRotationAngle()));
}

DoubleDescriptor BasicLattice2DItem::latticeLength1() const
{
    DoubleDescriptor d(getItem(P_LATTICE_LENGTH1), Unit::nanometer);
    d.tooltip = "Length of first lattice vector";
    return d;
}

void BasicLattice2DItem::setLatticeLength1(const double length1)
{
    setItemValue(P_LATTICE_LENGTH1, length1);
}

DoubleDescriptor BasicLattice2DItem::latticeLength2() const
{
    DoubleDescriptor d(getItem(P_LATTICE_LENGTH2), Unit::nanometer);
    d.tooltip = "Length of second lattice vector";
    return d;
}

void BasicLattice2DItem::setLatticeLength2(const double length2)
{
    setItemValue(P_LATTICE_LENGTH2, length2);
}

DoubleDescriptor BasicLattice2DItem::latticeAngle() const
{
    DoubleDescriptor d(getItem(P_LATTICE_ANGLE), Unit::degree);
    d.tooltip = "Angle between lattice vectors";
    return d;
}

void BasicLattice2DItem::setLatticeAngle(const double angle)
{
    setItemValue(P_LATTICE_ANGLE, angle);
}

// --------------------------------------------------------------------------------------------- //

SquareLattice2DItem::SquareLattice2DItem() : Lattice2DItem(M_TYPE)
{
    addProperty(P_LATTICE_LENGTH, 20.0)
        ->setToolTip("Length of first and second lattice vectors in nanometers");
}

std::unique_ptr<Lattice2D> SquareLattice2DItem::createLattice() const
{
    return std::make_unique<SquareLattice2D>(getItemValue(P_LATTICE_LENGTH).toDouble(),
                                             Units::deg2rad(latticeRotationAngle()));
}

DoubleDescriptor SquareLattice2DItem::latticeLength() const
{
    DoubleDescriptor d(getItem(P_LATTICE_LENGTH), Unit::nanometer);
    d.tooltip = "Length of first and second lattice vectors";
    return d;
}

void SquareLattice2DItem::setLatticeLength(const double length)
{
    setItemValue(P_LATTICE_LENGTH, length);
}

// --------------------------------------------------------------------------------------------- //

HexagonalLattice2DItem::HexagonalLattice2DItem() : Lattice2DItem(M_TYPE)
{
    addProperty(P_LATTICE_LENGTH, 20.0)
        ->setToolTip("Length of first and second lattice vectors in nanometers");
}

std::unique_ptr<Lattice2D> HexagonalLattice2DItem::createLattice() const
{
    return std::make_unique<HexagonalLattice2D>(getItemValue(P_LATTICE_LENGTH).toDouble(),
                                                Units::deg2rad(latticeRotationAngle()));
}

DoubleDescriptor HexagonalLattice2DItem::latticeLength() const
{
    DoubleDescriptor d(getItem(P_LATTICE_LENGTH), Unit::nanometer);
    d.tooltip = "Length of first and second lattice vectors";
    return d;
}

void HexagonalLattice2DItem::setLatticeLength(const double length)
{
    setItemValue(P_LATTICE_LENGTH, length);
}
