//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/ParticleCoreShellItem.cpp
//! @brief     Implements class ParticleCoreShellItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Material/MaterialModel.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Session/SessionItemUtils.h"
#include "GUI/Model/Session/SessionModel.h"
#include "GUI/Model/Types/DoubleDescriptor.h"
#include "GUI/Model/Types/VectorItem.h"
#include "GUI/Util/Error.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Particle/ParticleCoreShell.h"

namespace {

const QString abundance_tooltip = "Proportion of this type of particles normalized to the \n"
                                  "total number of particles in the layout";

const QString position_tooltip = "Relative position of the particle's reference point \n"
                                 "in the coordinate system of the parent (nm)";

} // namespace

ParticleCoreShellItem::ParticleCoreShellItem()
    : SessionItem(M_TYPE), ItemWithParticles(M_TYPE, abundance_tooltip, position_tooltip)
{
    registerTag(T_CORE, 0, 1, {ParticleItem::M_TYPE});
    registerTag(T_SHELL, 0, 1, {ParticleItem::M_TYPE});
}

std::unique_ptr<ParticleCoreShell> ParticleCoreShellItem::createParticleCoreShell() const
{
    std::unique_ptr<Particle> P_core{};
    std::unique_ptr<Particle> P_shell{};
    auto* core_item = dynamic_cast<ParticleItem*>(getItem(T_CORE));
    if (core_item)
        P_core = core_item->createParticle();
    auto* shell_item = dynamic_cast<ParticleItem*>(getItem(T_SHELL));
    if (shell_item)
        P_shell = shell_item->createParticle();
    if (!P_core || !P_shell)
        throw Error("ParticleCoreShellItem::createParticleCoreShell -> Error. Either "
                    "core or shell particle is undefined.");
    auto P_coreshell = std::make_unique<ParticleCoreShell>(*P_shell, *P_core);
    P_coreshell->setAbundance(abundance());
    setTransformationInfo(P_coreshell.get());
    return P_coreshell;
}

ParticleItem* ParticleCoreShellItem::core() const
{
    return dynamic_cast<ParticleItem*>(getItem(T_CORE));
}

void ParticleCoreShellItem::clearCore()
{
    if (core() != nullptr)
        model()->removeItem(core());
}

ParticleItem* ParticleCoreShellItem::createCore(MaterialModel* materials)
{
    auto* p = model()->insertItem<ParticleItem>(this, -1, T_CORE);
    p->setMaterialModel(materials);
    p->setMaterial(materials->defaultMaterial());
    return p;
}

bool ParticleCoreShellItem::isCoreTagName(const QString& name)
{
    return T_CORE == name;
}

ParticleItem* ParticleCoreShellItem::shell() const
{
    return dynamic_cast<ParticleItem*>(getItem(T_SHELL));
}

void ParticleCoreShellItem::clearShell()
{
    if (shell() != nullptr)
        model()->removeItem(shell());
}

ParticleItem* ParticleCoreShellItem::createShell(MaterialModel* materials)
{
    auto* p = model()->insertItem<ParticleItem>(this, -1, T_SHELL);
    p->setMaterialModel(materials);
    p->setMaterial(materials->defaultMaterial());

    // position is not used for shell item
    p->setPosition({});
    p->positionItem()->setEnabled(false); // legacy for SessionModel; can be removed after migration

    return p;
}

bool ParticleCoreShellItem::isShellTagName(const QString& name)
{
    return T_SHELL == name;
}
