//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/SampleValidator.cpp
//! @brief     Implements class SampleValidator
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/SampleValidator.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"

void SampleValidator::iterateItems(const SessionItem* parentItem)
{
    for (const SessionItem* child : parentItem->children()) {
        validateItem(child);
        iterateItems(child);
    }
}

void SampleValidator::validateItem(const SessionItem* item)
{
    if (const auto* p = dynamic_cast<const ParticleCoreShellItem*>(item)) {
        if (!p->core())
            addMessage("Core/shell particle doesn't have core defined.");
        if (!p->shell())
            addMessage("Core/shell particle doesn't have shell defined.");
    } else if (const auto* p = dynamic_cast<const ParticleCompositionItem*>(item))
        if (p->particles().isEmpty())
            addMessage("Particle composition doesn't have any particles.");
}

void SampleValidator::addMessage(const QString& m)
{
    m_messages += QString("* ") + m + "\n";
}

bool SampleValidator::isValidMultiLayer(const MultiLayerItem* multilayer)
{
    m_messages.clear();

    QVector<LayerItem*> layers = multilayer->layers();

    if (layers.isEmpty())
        addMessage("MultiLayer should contain at least one layer.");
    if (layers.size() == 1)
        if (layers.front()->layouts().isEmpty())
            addMessage("The single layer in your MultiLayer should contain a particle layout.");

    for (auto* layer : layers)
        for (auto* layout : layer->layouts()) {
            if (layout->particles().isEmpty())
                addMessage("Particle layout doesn't contain any particles.");
            for (auto* particle : layout->particles()) {
                validateItem(particle);
                iterateItems(particle);
            }
        }

    if (!m_messages.isEmpty())
        m_messages = "Can't setup DWBA simulation for given MultiLayer.\n" + m_messages;

    return m_messages.isEmpty();
}
