//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Session/SimulationOptionsItem.cpp
//! @brief     Defines class SimulationOptionsItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Session/SimulationOptionsItem.h"
#include "GUI/Model/Session/SessionXML.h"
#include "GUI/Util/DeserializationException.h"
#include <QXmlStreamWriter>
#include <thread>

namespace {

namespace Tags {

const QString RunImmediately("RunImmediately");
const QString NumberOfThreads("NumberOfThreads");
const QString Analytical("Analytical");
const QString NumberOfMonteCarloPoints("NumberOfMonteCarloPoints");
const QString UseAverageMaterials("UseAverageMaterials");
const QString IncludeSpecularPeak("IncludeSpecularPeak");

} // namespace Tags

} // namespace

using namespace GUI::Session::XML;

SimulationOptionsItem::SimulationOptionsItem()
{
    m_runImmediately = true;
    m_numberOfThreads = std::thread::hardware_concurrency();
    m_computationMethodAnalytical = true;
    m_numberOfMonteCarloPoints = 100;
    m_useAverageMaterials = false;
    m_includeSpecularPeak = false;
}

unsigned SimulationOptionsItem::numberOfThreads() const
{
    return m_numberOfThreads;
}

void SimulationOptionsItem::setNumberOfThreads(unsigned number)
{
    m_numberOfThreads = number;
}

bool SimulationOptionsItem::runImmediately() const
{
    return m_runImmediately;
}

bool SimulationOptionsItem::runInBackground() const
{
    return !runImmediately();
}

void SimulationOptionsItem::setRunImmediately(bool runImmediately)
{
    m_runImmediately = runImmediately;
}

void SimulationOptionsItem::setUseMonteCarloIntegration(unsigned numberOfPoints)
{
    m_computationMethodAnalytical = false;
    m_numberOfMonteCarloPoints = numberOfPoints;
}

void SimulationOptionsItem::setUseAnalytical()
{
    m_computationMethodAnalytical = true;
}

bool SimulationOptionsItem::useMonteCarloIntegration() const
{
    return !useAnalytical();
}

bool SimulationOptionsItem::useAnalytical() const
{
    return m_computationMethodAnalytical;
}

unsigned SimulationOptionsItem::numberOfMonteCarloPoints() const
{
    return m_numberOfMonteCarloPoints;
}

void SimulationOptionsItem::setUseAverageMaterials(bool useAverageMaterials)
{
    m_useAverageMaterials = useAverageMaterials;
}

bool SimulationOptionsItem::useAverageMaterials() const
{
    return m_useAverageMaterials;
}

void SimulationOptionsItem::setIncludeSpecularPeak(bool includeSpecularPeak)
{
    m_includeSpecularPeak = includeSpecularPeak;
}

bool SimulationOptionsItem::includeSpecularPeak() const
{
    return m_includeSpecularPeak;
}

void SimulationOptionsItem::writeContentTo(QXmlStreamWriter* writer) const
{
    writeAttribute(writer, GUI::Session::XML::Version, int(1));
    writeAttribute(writer, Tags::RunImmediately, m_runImmediately);
    writeAttribute(writer, Tags::NumberOfThreads, m_numberOfThreads);
    writeAttribute(writer, Tags::Analytical, m_computationMethodAnalytical);
    writeAttribute(writer, Tags::NumberOfMonteCarloPoints, m_numberOfMonteCarloPoints);
    writeAttribute(writer, Tags::UseAverageMaterials, m_useAverageMaterials);
    writeAttribute(writer, Tags::IncludeSpecularPeak, m_includeSpecularPeak);
}

void SimulationOptionsItem::readContentFrom(QXmlStreamReader* reader)
{
    const int version = reader->attributes().value(GUI::Session::XML::Version).toInt();

    if (version < 1)
        throw DeserializationException::tooOld();

    if (version > 1)
        throw DeserializationException::tooNew();

    m_runImmediately = readBoolAttribute(reader, Tags::RunImmediately);
    m_numberOfThreads = readUIntAttribute(reader, Tags::NumberOfThreads);
    m_computationMethodAnalytical = readBoolAttribute(reader, Tags::Analytical);
    m_numberOfMonteCarloPoints = readUIntAttribute(reader, Tags::NumberOfMonteCarloPoints);
    m_useAverageMaterials = readBoolAttribute(reader, Tags::UseAverageMaterials);
    m_includeSpecularPeak = readBoolAttribute(reader, Tags::IncludeSpecularPeak);
}
