//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/EnvironmentEditor.cpp
//! @brief     Implements class EnvironmentEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/EnvironmentEditor.h"
#include "GUI/Model/Instrument/BackgroundItems.h"
#include "GUI/Model/Instrument/InstrumentItems.h"

#include <QComboBox>
#include <QDoubleSpinBox>
#include <QFormLayout>
#include <QGroupBox>
#include <QStackedLayout>

//==================================================================================================
// ConstantBackgroundForm
//==================================================================================================

class ConstantBackgroundForm : public QWidget {
public:
    ConstantBackgroundForm(QWidget* parent = nullptr, Qt::WindowFlags f = Qt::WindowFlags());
    void setupBackground(ConstantBackgroundItem* item);

private slots:
    void onValueChanged(double value);

private:
    QDoubleSpinBox* m_spinbox;
    ConstantBackgroundItem* m_item;
};

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

ConstantBackgroundForm::ConstantBackgroundForm(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    m_spinbox = new QDoubleSpinBox(this);
    connect(m_spinbox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &ConstantBackgroundForm::onValueChanged);
    m_spinbox->setDecimals(3);
    m_spinbox->setRange(0, std::numeric_limits<double>::max());
    m_spinbox->setSingleStep(0.01);
    m_spinbox->setToolTip("Constant background value [counts/pixel]");
    layout->addRow("Background value:", m_spinbox);
}

void ConstantBackgroundForm::setupBackground(ConstantBackgroundItem* item)
{
    m_item = item;
    setEnabled(m_item);
    if (m_item)
        m_spinbox->setValue(m_item->backgroundValue());
    else
        m_spinbox->setValue(0);
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void ConstantBackgroundForm::onValueChanged(double value)
{
    if (m_item)
        m_item->setBackgroundValue(value);
}

//==================================================================================================
// EnvironmentEditor
//==================================================================================================

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

EnvironmentEditor::EnvironmentEditor(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f), m_item(nullptr)
{
    auto* layout = new QVBoxLayout(this);

    auto* gbox = new QGroupBox("Background", this);
    layout->addWidget(gbox);
    auto* form = new QFormLayout(gbox);

    m_combo = new QComboBox(gbox);
    form->addRow("Type:", m_combo);

    m_stack = new QStackedLayout();
    form->addRow(m_stack);

    m_combo->addItem("None", NoBackgroundItem::M_TYPE);
    m_combo->addItem("Constant", ConstantBackgroundItem::M_TYPE);
    m_combo->addItem("Poisson noise", PoissonNoiseBackgroundItem::M_TYPE);
    connect(m_combo, qOverload<int>(&QComboBox::currentIndexChanged), this,
            &EnvironmentEditor::onBackgroundSelected);

    m_blankForm = new QWidget(gbox);
    m_stack->addWidget(m_blankForm);

    m_constBgForm = new ConstantBackgroundForm(gbox);
    m_stack->addWidget(m_constBgForm);
}

void EnvironmentEditor::setItem(InstrumentItem* item)
{
    m_item = item;
    setEnabled(m_item);
    if (m_item) {
        int idx = m_combo->findData(m_item->backgroundItem()->modelType());
        ASSERT(idx != -1);
        if (idx != m_combo->currentIndex())
            m_combo->setCurrentIndex(idx);
        else if (m_item->backgroundItem()->hasModelType<ConstantBackgroundItem>())
            onBackgroundSelected(idx); // trigger update manually
    }
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void EnvironmentEditor::onBackgroundSelected(int index)
{
    if (m_item) {
        const QString bgtype = m_combo->itemData(index).toString();
        if (bgtype == ConstantBackgroundItem::M_TYPE) {
            m_constBgForm->setupBackground(m_item->setBackgroundType<ConstantBackgroundItem>());
            m_stack->setCurrentWidget(m_constBgForm);
        } else if (bgtype == NoBackgroundItem::M_TYPE) {
            m_item->setBackgroundType<NoBackgroundItem>();
            m_stack->setCurrentWidget(m_blankForm);
        } else if (bgtype == PoissonNoiseBackgroundItem::M_TYPE) {
            m_item->setBackgroundType<PoissonNoiseBackgroundItem>();
            m_stack->setCurrentWidget(m_blankForm);
        } else
            ASSERT(false);
    }
}
