//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Particle/MesoCrystal.cpp
//! @brief     Implements class MesoCrystal.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Particle/MesoCrystal.h"
#include "Base/Util/Assert.h"
#include "Sample/Particle/Crystal.h"
#include "Sample/Particle/SlicedParticle.h"
#include "Sample/Scattering/FormFactorDecoratorPositionFactor.h"
#include "Sample/Scattering/FormFactorDecoratorRotation.h"
#include "Sample/Scattering/Rotations.h"

MesoCrystal::MesoCrystal(Crystal* particle_structure, IFormFactor* form_factor)
    : m_particle_structure(particle_structure), m_meso_form_factor(form_factor)
{
}

MesoCrystal::MesoCrystal(const Crystal& particle_structure, const IFormFactor& form_factor)
    : MesoCrystal(particle_structure.clone(), form_factor.clone())
{
}

MesoCrystal::~MesoCrystal() = default;

MesoCrystal* MesoCrystal::clone() const
{
    auto* p_result = new MesoCrystal(m_particle_structure->clone(), m_meso_form_factor->clone());
    p_result->setAbundance(m_abundance);
    if (m_rotation)
        p_result->setRotation(*m_rotation);
    p_result->setParticlePosition(m_position);
    return p_result;
}

SlicedParticle MesoCrystal::createSlicedParticle(const ZLimits& limits) const
{
    ASSERT(m_particle_structure && m_meso_form_factor);
    std::unique_ptr<IFormFactor> tem_ff(
        m_meso_form_factor->createSlicedFormFactor(limits, m_rotation.get(), m_position));
    std::unique_ptr<IFormFactor> total_ff(
        m_particle_structure->createTotalFormFactor(*tem_ff, m_rotation.get(), m_position));
    double meso_volume = m_meso_form_factor->volume();
    auto regions = m_particle_structure->admixtures();
    for (auto& region : regions)
        region.fraction *= meso_volume;
    SlicedParticle result;
    result.sliced_ff = std::move(total_ff);
    result.admixtures = regions;
    return result;
}

std::vector<const INode*> MesoCrystal::getChildren() const
{
    return std::vector<const INode*>()
           << IParticle::getChildren() << m_particle_structure << m_meso_form_factor;
}

const Crystal& MesoCrystal::particleStructure() const
{
    ASSERT(m_particle_structure);
    return *m_particle_structure;
}
