//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Particle/Particle.cpp
//! @brief     Implements class Particle.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Particle/Particle.h"
#include "Base/Util/Assert.h"
#include "Base/Vector/Transform3D.h"
#include "Sample/Material/MaterialFactoryFuncs.h"
#include "Sample/Particle/SlicedParticle.h"
#include "Sample/Scattering/FormFactorDecoratorMaterial.h"
#include "Sample/Scattering/FormFactorDecoratorPositionFactor.h"
#include "Sample/Scattering/Rotations.h"

Particle::~Particle() = default;

Particle::Particle(Material material, const IFormFactor& form_factor)
    : m_material(std::move(material)), m_form_factor(form_factor.clone())
{
}

Particle::Particle(Material material, const IFormFactor& form_factor, const IRotation& rotation)
    : Particle(material, form_factor)
{
    setRotation(rotation);
}

Particle* Particle::clone() const
{
    ASSERT(m_form_factor);
    auto* p_result = new Particle(m_material, *m_form_factor);
    p_result->setAbundance(m_abundance);
    if (m_rotation)
        p_result->setRotation(*m_rotation);
    p_result->setParticlePosition(m_position);

    return p_result;
}

SlicedParticle Particle::createSlicedParticle(const ZLimits& limits) const
{
    ASSERT(m_form_factor);
    std::unique_ptr<IFormFactor> sliced_raw_ff(
        m_form_factor->createSlicedFormFactor(limits, m_rotation.get(), m_position));
    if (!sliced_raw_ff)
        return {};
    auto sliced_ff = std::make_unique<FormFactorDecoratorMaterial>(*sliced_raw_ff);
    double volume = sliced_raw_ff->volume();
    Material transformed_material(
        m_rotation ? m_material.rotatedMaterial(m_rotation->getTransform3D()) : m_material);
    sliced_ff->setMaterial(transformed_material);
    return {std::move(sliced_ff), {{{volume, transformed_material}}}};
}

std::vector<const INode*> Particle::getChildren() const
{
    return std::vector<const INode*>() << IParticle::getChildren() << m_form_factor;
}
