//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Scattering/IFormFactor.cpp
//! @brief     Implements interface class IFormFactor.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Scattering/IFormFactor.h"
#include "Base/Util/PyFmt.h"
#include "Base/Util/StringUtils.h"
#include "Base/Vector/WavevectorInfo.h"
#include "Sample/Scattering/FormFactorDecoratorPositionFactor.h"
#include "Sample/Scattering/FormFactorDecoratorRotation.h"
#include "Sample/Scattering/Rotations.h"
#include <memory>
#include <utility>

namespace {

bool shapeIsContainedInLimits(const IFormFactor& formfactor, ZLimits limits,
                              const IRotation* rotation, R3 translation)
{
    double zbottom = formfactor.bottomZ(rotation) + translation.z();
    double ztop = formfactor.topZ(rotation) + translation.z();

    return limits.zBottom() <= zbottom && ztop <= limits.zTop();
}

bool shapeOutsideLimits(const IFormFactor& formfactor, ZLimits limits,
                        const IRotation* rotation, R3 translation)
{
    double zbottom = formfactor.bottomZ(rotation) + translation.z();
    double ztop = formfactor.topZ(rotation) + translation.z();

    return ztop <= limits.zBottom() || zbottom >= limits.zTop();
}

} // namespace


IFormFactor::IFormFactor(const NodeMeta& meta, const std::vector<double>& PValues)
    : ISampleNode(meta, PValues), m_nodeMeta(meta)
{
}

std::string IFormFactor::shapeName() const
{
    if (className().substr(0, 10) == "FormFactor")
        return className().substr(10);
    return className();
}

IFormFactor* IFormFactor::createSlicedFormFactor(ZLimits limits, const IRotation* rot,
                                                 R3 translation) const
{
    if (shapeIsContainedInLimits(*this, limits, rot, translation))
        return createTransformedFormFactor(*this, rot, translation);
    if (shapeOutsideLimits(*this, limits, rot, translation))
        return nullptr;
    if (canSliceAnalytically(rot))
        return sliceFormFactor(limits, rot, translation);
    throw std::runtime_error("Slicing of " + className()
                             + " not supported for the given rotation!");
}

Eigen::Matrix2cd IFormFactor::thePolFF(const WavevectorInfo&) const
{
    // Throws to prevent unanticipated behaviour
    throw std::runtime_error("IFormFactor::evaluatePol: is not implemented by default");
}

std::string IFormFactor::pythonConstructor() const
{
    std::vector<std::pair<double, std::string>> arguments;
    for (size_t i = 0; i < m_nodeMeta.paraMeta.size(); i++)
        arguments.emplace_back(m_P[i], m_nodeMeta.paraMeta[i].unit);

    return Py::Fmt::printFunction(className(), arguments);
}

double IFormFactor::volume() const
{
    auto zero_wavevectors = WavevectorInfo::makeZeroQ();
    return std::abs(theFF(zero_wavevectors));
}

bool IFormFactor::canSliceAnalytically(const IRotation*) const
{
    return false;
}

IFormFactor* IFormFactor::sliceFormFactor(ZLimits, const IRotation*, R3) const
{
    throw std::runtime_error(className() + "::sliceFormFactor error: not implemented!");
}

IFormFactor* IFormFactor::createTransformedFormFactor(const IFormFactor& formfactor,
                                                      const IRotation* rot, R3 translation)
{
    std::unique_ptr<IFormFactor> P_fftemp, P_result;
    if (rot && !rot->isIdentity())
        P_fftemp = std::make_unique<FormFactorDecoratorRotation>(formfactor, *rot);
    else
        P_fftemp.reset(formfactor.clone());
    if (translation != R3())
        P_result = std::make_unique<FormFactorDecoratorPositionFactor>(*P_fftemp, translation);
    else
        std::swap(P_fftemp, P_result);
    return P_result.release();
}
